/*
 GB 64MB Mighty Flash Cart - Flash the cart using GBxCart RW
 Version: 1.0
 Author: Alex from insideGadgets (www.insidegadgets.com)
 Created: 17/12/2018
 Last Modified: 17/12/2018
 
 This program allows you to write your ROM file to the GB 64MB Mighty Flash Cart
 
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <time.h>

#ifdef _WIN32
#include <Windows.h>
#else
#include <unistd.h>
#endif

#include "setup.h" // See defines, variables, constants, functions here

int main(int argc, char **argv) {
	
	printf("insideGadgets - 64MB Mighty Flash Cart Flasher\n");
	printf("##############################################\n");
	
	uint8_t useFastWrite = 0;
	
	// Check arguments
	if (argc >= 3) {
		read_config();
		
		// Use fast write
		if (atoi(argv[1]) == 1) {
			useFastWrite = 1;
			printf("Using fast writing\n");
		}
		
		// Open COM port
		if (com_test_port() == 0) {
			printf("Device not connected and couldn't be auto detected\n");
			read_one_letter();
			return 1;
		}
		printf("Connected on COM port: %i\n", cport_nr+1);
		
		// Break out of any existing functions on ATmega
		set_mode('0');
		
		// Get cartridge mode - Gameboy or Gameboy Advance
		cartridgeMode = request_value(CART_MODE);
		
		// Get firmware version
		gbxcartFirmwareVersion = request_value(READ_FIRMWARE_VERSION);
		
		// Get PCB version
		gbxcartPcbVersion = request_value(READ_PCB_VERSION);
		
		if (gbxcartPcbVersion == PCB_1_0) {
			printf("\nPCB v1.0 is not supported for this function.");
			read_one_letter();
			return 1;
		}
		
		if (gbxcartFirmwareVersion <= 11) {
			printf("Firmware R12 or higher is required for this functionality.\n");
			read_one_letter();
			return 1;
		}
		
		
		// Separate the file name
		char filenameOnly[100];
		char filename[254];
		strncpy(filename, argv[2], 253);
		
		char *token = strtok(filename, "\\");
		while (token != NULL) {
			strncpy(filenameOnly, token, 99);
			token = strtok(NULL, "\\");
		}
		
		printf("\n--- Write ROM to Flash Cart ---\n");
		
		uint32_t readBytes = 0;
		uint16_t sector = 0;
		long fileSize = 0;
		
		// Grab file size
		FILE *romFile = fopen(argv[2], "rb");
		if (romFile != NULL) {
			fseek(romFile, 0, SEEK_END);
			fileSize = ftell(romFile);
			fseek(romFile, 0, SEEK_SET);
		}
		else {
			printf("\n%s \nFile not found\n", argv[1]);
			read_one_letter();
			return 1;
		}
		
		
		printf("\nGoing to write to ROM (Flash cart) from %s\n", filenameOnly);
		
		// PCB v1.3 - Set 5V
		if (gbxcartPcbVersion == PCB_1_3) {
			set_mode(VOLTAGE_5V);
		}
		
		// Check file size
		if (fileSize > 0x4000000) {
			fclose(romFile);
			printf("\n%s \nFile size is larger than the available Flash cart space of 64 MByte\n", argv[1]);
			read_one_letter();
			return 1;
		}
		
		currAddr = 0x0000;
		endAddr = 0x7FFF;
		
		// Calculate banks/8MB blocks needed from ROM file size
		romBanks = fileSize / 16384;
		uint8_t romBlocks = (romBanks / 512) + 1;
		uint16_t romBanksCalc = fileSize / 16384;
		uint16_t romBanksTotal = romBanksCalc;
		if (romBlocks > 1) {
			romBanks = 512;
		}
		if (romBlocks == 0) {
			romBlocks = 1;
		}
		
		// Detect if save slots are being used by 
		uint8_t saveSlotsDetected = 0;
		uint32_t zeroCounter = 0;
		
		if (fileSize > 0x120000) { // File needs to be more than this for save slots to be active
			fseek(romFile, 0x20000, SEEK_SET); // First save slot
			
			for (uint16_t counter = 0; counter < 512; counter++) {
				uint8_t buffer[256];
				fread(&buffer, 1, 256, romFile);
				
				for (uint16_t z = 0; z < 256; z++) {
					if (buffer[z] == 0) {
						zeroCounter++;
					}
				}
			}
			
			if (zeroCounter == 0x20000) { // 131072 bytes of all 0's
				saveSlotsDetected = 1;
				printf("Save slots detected in ROM file\n");
			}
			fseek(romFile, 0x00, SEEK_SET);
		}
		
		// Flash Setup
		set_mode(GB_CART_MODE); // Gameboy mode
		gb_flash_pin_setup(WE_AS_WR_PIN); // WR pin
		gb_flash_program_setup(GB_FLASH_PROGRAM_AAA);// Flash program byte method
		
		printf("\n\nWriting to ROM (Flash cart) from %s\n", filenameOnly);
		printf("[             25%%             50%%             75%%            100%%]\n[");
		
		
		// Set bank as 0
		set_bank(0x2100, 0);
		set_bank(0x3000, 0);
		
		// Change to first 8MB block
		set_bank(0x1000, 0xD9); // Unlock multi-game mode
		set_bank(0x7000, 0xF0); // Set MBC5 mode and high 3 bits of bank value
		set_bank(0x6000, 0x00); // Bank
		set_bank(0x0000, 0xAA); // Turn off multi-game mode
		
		
		// Write ROM
		currAddr = 0x0000;
		for (uint8_t block = 0; block < romBlocks; block++) {			
			currAddr = 0x0000;
			sector = 0;
			
			for (uint16_t bank = 1; bank < romBanks; bank++) {				
				if (bank > 1) { currAddr = 0x4000; }
				
				// Set start address
				set_number(currAddr, SET_START_ADDRESS);
				
				// Read data
				while (currAddr < endAddr) {
					if (currAddr == 0x4000) { // Switch banks here just before the next bank
						set_bank(0x2100, bank);
						if (bank >= 256) {
							set_bank(0x3000, 1); // High bit
						}
					}
					
					// If save slots are present - Write the loader, skip writing to save slots and then write data after save slots
					if (saveSlotsDetected == 0 || (saveSlotsDetected == 1 && (readBytes < 0x20000 || readBytes >= 0x120000))) {
						if (readBytes % 0x20000 == 0) { // Erase sectors
							gb_flash_write_address_byte(0xAAA, 0xAA);
							gb_flash_write_address_byte(0x555, 0x55);
							gb_flash_write_address_byte(0xAAA, 0x80);
							gb_flash_write_address_byte(0xAAA, 0xAA);
							gb_flash_write_address_byte(0x555, 0x55);
							gb_flash_write_address_byte(0x4000, 0x30);
							
							wait_for_flash_sector_ff(currAddr);
							sector++;
							
							set_number(currAddr, SET_START_ADDRESS);
						}
						
						// Regular writing
						if (useFastWrite == 0) {
							com_write_bytes_from_file(GB_FLASH_WRITE_64BYTE, romFile, 64);
							com_wait_for_ack();
							currAddr += 64;
							readBytes += 64;
						}
						else {
							//printf("addr 0x%X\n", currAddr);
							com_write_bytes_from_file(GB_FLASH_WRITE_BUFFERED_32BYTE, romFile, 32);
							com_wait_for_ack();
							currAddr += 32;
							readBytes += 32;
						}
					}
					else { // Skip ROM
						fread(writeBuffer, 1, 64, romFile);
						currAddr += 64;
						readBytes += 64;
					}
					
					// Print progress
					print_progress_percent(readBytes, (romBanksTotal * 16384) / 64);
				}
			}
			
			// Calculate remaining rom banks
			if (romBanksCalc > 512) {
				romBanksCalc = romBanksCalc - 512;
				if (romBanksCalc <= 512) {
					romBanks = romBanksCalc;
				}
			}
			
			
			// Set bank as 0
			set_bank(0x2100, 0);
			set_bank(0x3000, 0);
			
			// Change to next 8MB block
			uint8_t nextBlock = 0xF2 + (block * 2);
			set_bank(0x1000, 0xD9); // Unlock multi-game mode
			set_bank(0x7000, nextBlock); // Set MBC5 mode and high 3 bits of bank value
			set_bank(0x6000, 0x00); // Bank
			set_bank(0x0000, 0xAA); // Turn off multi-game mode
		}
		
		
		// Set bank as 0
		set_bank(0x2100, 0);
		set_bank(0x3000, 0);
		
		// Change to first 8MB block
		set_bank(0x1000, 0xD9); // Unlock multi-game mode
		set_bank(0x7000, 0xF0); // Set MBC5 mode and high 3 bits of bank value
		set_bank(0x6000, 0x00); // Bank
		set_bank(0x0000, 0xAA); // Turn off multi-game mode
		
		printf("]");
		fclose(romFile);
	}
	else {
		printf("Please drag and drop your ROM file\n");
	}
	
	printf("\n");
	
	return 0;
}